#property copyright "PipDefender - by Lorenzo Borrelli - 2010"
#property link "ellebimail@gmail.com"

#include <stdlib.mqh>
#include <WinUser32.mqh>

/*   
   Boolean MQ4 values MUST be passed to C++ as integer 0/1 values (note: gm, sm and sfm in SetExternalParams are Bool values).
   Boolean values returned by C++ functions are returned as integer 0/1 values, these values could be normally used as MQ4 bool values
*/

#import "pipdefender.dll"
   void EndTicketCycle();
   void MarkAsSafe();
   int IsSafe();
   void SetEnvironment( int pm, double ts, int gm, int gl, int sm, double tsl, double ftp, double fsl, int sfm, double mhw, double mha, double sfl,double pv, double xsl);
   void SetOrderData( double ask, double bid, int tk, int ty, double op, double sl, double tp, /*double lo,*/ int ot);
   void UpdateOrder( double ask, double bid, double sl, double tp);
   int SafeMode( datetime currentTime );
   void StealthMode( int& returnCode[], double& stealthSl[], double& stealthTp[]);
   void ManageProfit( int& returnCode[], double& stealthSl[], double& stealthTp[]);
#import

// -------------------------------------------------------------------------------------------------------------------------------
// CONST
// -------------------------------------------------------------------------------------------------------------------------------
// PIPDEFENDER.DLL CONST
#define PIPD_FALSE 0                   // "boolean" value for FALSE returned from dll functions
#define PIPD_TRUE 1                    // "boolean" value for TRUE returned from dll functions
#define PIPD_MODIFY -1                 // command action: "modify order" (sl & tp are passed separately)
#define PIPD_CLOSE -2                  // command action: "close order"
#define PIPD_ABORT  -3                 // ABORT command sent from the dll
#define PIPD_NOTHING -19740204         // command action: "no action"
#define PIPD_D_NOTHING -1974.0204      // double "null" value

// Other constants
#define _INDENT  "\n          "
#define _INDENT2 "   "
#define _SEPARATOR "----------------------------------------------------------------------"
#define _ALL_OK 0                       // no exceptions, the EA continue to work normally
#define _ALERT -1                       // An "Alert" box will be shown, the EA suspend every other activity
#define _ABORT -2                       // An "Alert" box will be shown, the EA abort and will be removed from its chart
#define _SUSPEND -3                     // Special value used to avoid multiple Alert boxes
#define _NOTFOUND -4
#define _CLOSE_MAXTRY 10                // maximum number of consecutive retry for a bad OrderClose operation
#define _MODIFY_MAXTRY 10               // maximum number of consecutive retry for the main retry cycle
#define _BUSY_MAXTRY 3                  // maximum number of consecutive retry for a busy trade context (OrderModify)
#define _BUSY_SLEEPTIME 500             // default sleeptime (milliseconds) for the sleep function (when the trade context is busy)

// -------------------------------------------------------------------------------------------------------------------------------
// EXTERNAL PARAMETERS
// -------------------------------------------------------------------------------------------------------------------------------
extern string VERSION = "1.0beta";
extern bool FIBONACCI_MODE = false;    // true for Fibonacci mode on
// --- Standard mode parameters
extern string _____Standard_mode_parameters = _SEPARATOR;
extern double TRAILING_STOP = 4.0;     // PIP
extern bool GREEDY_MODE = true;        // Enable or disable the greedy mode
extern int GREEDY_LEVEL = 2;           // number of TS interval to be managed with the classic trailing top algorithm
// --- Stealth mode parameters
extern string _____Stealth_mode_parameters = _SEPARATOR;
extern bool STEALTH_MODE = true;       // Enable or disable the internal SL management
extern double TRUE_SL = 40.0;          // PIP
extern double FAKE_TP = 50.0;          // PIP
extern double FAKE_SL = 150.0;         // PIP
// --- Safe mode parameters string
extern string _____Safe_mode_parameters = _SEPARATOR;
extern bool SAFE_MODE = true;          // Enable or disable the safe mode
extern double MAX_HOURS_WARNING = 8.0;
extern double MAX_HOURS_ALARM = 16.0;
extern double SAFE_LOSS = 4.0;         // Max "safe" PIP loss after MAX_HOURS
extern string _____Other_parameters = _SEPARATOR;
extern double CLOSE_SLIPPAGE = 3.0;    // PIP
 
// -------------------------------------------------------------------------------------------------------------------------------
// GLOBAL VARIABLES
// -------------------------------------------------------------------------------------------------------------------------------
static int EXCEPTION = _ALL_OK ;
double extraPip;
double POINT_VALUE;         // point value ( eg. 0.00010 for a 5 digits account )
double X_STOPLEVEL;         // stop level in PIP form for the current cross
int O_TICKET;               // current selected order ticket
int O_TYPE;                 // order type: OP_BUY or OP_SELL
//double O_LOTS;              // size...
double O_OPENPRICE;         // open price
double O_STOPLOSS;
double O_TAKEPROFIT;
double M_ASK;               // current market Ask price
double M_BID;               // current market Bid price

// -------------------------------------------------------------------------------------------------------------------------------
// INIT FUNCTION
// -------------------------------------------------------------------------------------------------------------------------------

int init() {
   // Forcing the load of the stdlib.mqh
   ErrorDescription( GetLastError() );

   // Calculation of Global Parameters
   if ( Equal( Point, 0.00001 ) )
      POINT_VALUE=0.0001;
   else
      if ( Equal( Point, 0.001 ) )
         POINT_VALUE=0.01;
      else
         POINT_VALUE = Point;

   X_STOPLEVEL = MarketInfo( Symbol(), MODE_STOPLEVEL ) * 0.1;    // stop level in PIP

   // External Parameters check and print the info panel
   CheckParameters(); 
   //DisplayHeader();
   
   // =====> to Pipdefender.dll  
   SetEnvironment( FIBONACCI_MODE, TRAILING_STOP, GREEDY_MODE, GREEDY_LEVEL, STEALTH_MODE, TRUE_SL, FAKE_TP, FAKE_SL, SAFE_MODE, MAX_HOURS_WARNING, MAX_HOURS_ALARM, SAFE_LOSS, POINT_VALUE, X_STOPLEVEL);

   return( 0 );
}

// -------------------------------------------------------------------------------------------------------------------------------
// DEINIT FUNCTION
// -------------------------------------------------------------------------------------------------------------------------------

int deinit() {
   return( 0 );
}

// -------------------------------------------------------------------------------------------------------------------------------
// START FUNCTION
// -------------------------------------------------------------------------------------------------------------------------------

int start() {
int O_TOTAL;              // Actual size of the order's pool; Const number inside each start() call
double extraPip;
bool result;
int retCode[1];           // StealthMode & Manage return code   
double retSl[1];          // suggested stop loss level or PIPD_D_NOTHING
double retTp[1];          // suggested take profit or PIPD_D_NOTHING

if ( EXCEPTION == _SUSPEND ) return ( 0 );

if ( EXCEPTION == _ABORT ) {
   EXCEPTION = _SUSPEND;
   Alert( "PipDefender Aborted! The EA will be removed from its chart." );
   // this will remove expert from chart
   PostMessageA( WindowHandle( Symbol(), Period()), WM_COMMAND, 33050, 0);
   return( 0 );
}

if ( EXCEPTION == _ALERT ) {
   EXCEPTION = _SUSPEND;
   Alert( "PipDefender Error! The EA will be disabled." );
   return( 0 );
}

O_TOTAL = OrdersTotal();
if ( O_TOTAL == 0 ) {
   if ( extraPip != 0 ) {
      Cout ("No buy or sell " + Symbol() + " open orders to take care of...");
      CoutF ( "... but I've already saved " + DoubleOneDigitToString( extraPip ) + " pip" );
   }
   else
      CoutF ("No buy or sell " + Symbol() + " open orders to take care of...");  
   return ( 0 );
}

// Show the header information bar
ShowDisplay();

for ( int i = 0; i < O_TOTAL; i++ ) {
   // SelectOrder return two different things:
   // 1) the ticket number of a Symbol() BUY/SELL order
   // 2) the _NOTFOUND flag (error)
   O_TICKET = SelectOrder( i );
   if ( O_TICKET != _NOTFOUND ) {
      // Order-constant data            
      O_TYPE = OrderType();
      O_OPENPRICE = OrderOpenPrice();
      //O_LOTS = OrderLots();
      O_STOPLOSS = OrderStopLoss();
      O_TAKEPROFIT = OrderTakeProfit();                          
      M_BID = MarketInfo( Symbol(), MODE_BID );
      M_ASK = MarketInfo( Symbol(), MODE_ASK );
      // =====> to Pipdefender.dll
      SetOrderData( M_ASK, M_BID, O_TICKET, O_TYPE, O_OPENPRICE, O_STOPLOSS, O_TAKEPROFIT, /*O_LOTS,*/ OrderOpenTime());
      
      // ------ SAFE MODE MANAGEMENT --------- 
      // <===== from Pipdefender.dll     
      if ( SafeMode( TimeCurrent() ) == PIPD_CLOSE ) {      
         if ( CloseOrder() )
            Print( "====>       SAFE MODE ", Symbol(), " #" + O_TICKET + " [" + OrderTypeToString() + "] | open=" + DoubleDigitsToString( O_OPENPRICE ) + " | " + OrderPriceName() + "=" + DoubleDigitsToString( OrderPrice() )  + " | profit= " + ShowProfit() );      
      }
      // -------------------------------------

      // ------ STEALTH MODE MANAGEMENT ------
      // <===== from Pipdefender.dll     
      StealthMode( retCode, retSl, retTp);     
      switch ( retCode[0] ) {
         case PIPD_CLOSE:
            if ( CloseOrder() )
               Print( "====>       STOP LOSS ", Symbol(), " #" + O_TICKET + " [" + OrderTypeToString() + "] | open=" + DoubleDigitsToString( O_OPENPRICE ) + " | " + OrderPriceName() + "=" + DoubleDigitsToString( OrderPrice() ) + " | loss= " + ShowProfit() );
            continue;
         case PIPD_MODIFY:
            // If ModifyOrder fail we retry on next tick
            ModifyOrder( retSl[0], retTp[0] );   
            break;
         default: // PIPD_NOTHING
            // nothing to do...
            break;
      }       
      // -------------------------------------

      LineFeed();
      Cout("Order= " + O_TICKET + " [" + OrderTypeToString() + "]");      
      // Show additional infos on a already modified order
      // <===== from Pipdefender.dll     
      if ( IsSafe() == PIPD_TRUE ) {        
         Cout( _INDENT2 +  "Already saved " + DoubleOneDigitToString( MathAbs( O_STOPLOSS - O_OPENPRICE ) / POINT_VALUE ) + " pip");         
         Cout( _INDENT2 +  "Current profit= " + ShowProfit() );
      }

      // ------ PROFIT MANAGEMENT ------------      
      for ( int n = 1; n <= _MODIFY_MAXTRY; n++ ) {                                                                                    
            // <===== from Pipdefender.dll
            ManageProfit ( retCode, retSl, retTp );                 
            if ( retCode[0] == PIPD_MODIFY )
               result = ModifyOrderAndStore( retSl[0], retTp[0] );                           
            else
               if ( retCode[0] == PIPD_ABORT ) {
                  EXCEPTION = _ABORT;
                  Alert(" PipDefender beta expired! ");
                  return;
               }                  
               else
                  // retCode[0] == PIPD_NOTHING
                  break;   // exit for
                  
            if ( result ) {                            
               Print( "====>       ", Symbol(), " #" + O_TICKET + " [" + OrderTypeToString() + "] | open=" + DoubleDigitsToString( O_OPENPRICE ) + " | " + OrderPriceName() + "=" + DoubleDigitsToString( OrderPrice() )  + " | safe profit= " + DoubleOneDigitToString( MathAbs( O_STOPLOSS - O_OPENPRICE ) / POINT_VALUE ) + " PIP" );            
               break;   // exit for
            }
            else {   // ModifyOrderAndStore failure: update & retry          
               M_BID = MarketInfo( Symbol(), MODE_BID );
               M_ASK = MarketInfo( Symbol(), MODE_ASK );
               //O_STOPLOSS = OrderStopLoss();
               //O_TAKEPROFIT = OrderTakeProfit();
               // =====> to Pipdefender.dll
               UpdateOrder( M_ASK, M_BID , O_STOPLOSS, O_TAKEPROFIT );            
               continue;
            }
         }
      // -------------------------------------
            
   } // end if 
} // end for
// Update the ticket cache in PipDefender.dll
// =====> to Pipdefender.dll
EndTicketCycle();
return( 0 );
}

// -------------------------------------------------------------------------------------------------------------------------------
// NUMERICAL FUNCTIONS FOR SAFE NUMBER COMPARISON
// -------------------------------------------------------------------------------------------------------------------------------

// Return the normalized value with "Digits" digits
double NormalizeDigits( double value ) {   
   return ( NormalizeDouble( value, Digits ) );
}

// Format the number with "Digits" digits
string DoubleDigitsToString( double value ) {
   return ( DoubleToStr( value, Digits ) );
}

// Format the number with 1 digit
string DoubleOneDigitToString( double price ) {
   return ( DoubleToStr( price, 1 ) );
}

bool Equal( double number1, double number2 ) {
   return ( NormalizeDouble( number1 - number2, 8 ) == 0 );
}

bool IsLesserEqual(double number1, double number2) {
   return ( NormalizeDouble( number1 - number2, 8 ) <= 0 );
}

// -------------------------------------------------------------------------------------------------------------------------------
// ORDER SELECTING & MANAGEMENT FUNCTIONS
// -------------------------------------------------------------------------------------------------------------------------------

// Select the order with the passed position only if it is a BUY/SELL and Symbol order, otherwise return NOT FOUND
int SelectOrder(int pos) {
   if ( OrderSelect( pos, SELECT_BY_POS, MODE_TRADES ) )
      if ( OrderSymbol() == Symbol() )
         if ( ( OrderType() == OP_BUY ) || ( OrderType() == OP_SELL ) )
            return ( OrderTicket() );
   return ( _NOTFOUND );
}

// Return TRUE only if the order was correctly modified, FALSE otherwise
bool ModifyOrderAndStore( double newSl, double newTp ) {
   double oldSl = O_STOPLOSS;
   bool result;
   
   result = ModifyOrder( newSl, newTp );   
   if ( result ) {
      // SUCCESS!!                       
      // if the order is already known the previous OrderModify has just granted **additional** extra pip of profit
      // if the order is pretty new the EA store its ticked number for future recognition                        
      // <===== from Pipdefender.dll     
      if ( IsSafe() == PIPD_TRUE )
         extraPip = extraPip + ( MathAbs( newSl - oldSl ) / POINT_VALUE );
      else {
         MarkAsSafe();
         Print("====>       ", Symbol(), " #", O_TICKET, " is now taken under PipDefender protection!");
      }
   }   
   return ( result );
}

// Return TRUE only if the order was correctly modified, FALSE otherwise
bool ModifyOrder( double newSl, double newTp ) {
   bool result;
   int err;   
      
   if ( IsTradeContextBusy() || !IsTradeAllowed() )   
      for ( int i = 1; i <= _BUSY_MAXTRY; i++ ) {         
         Sleep ( _BUSY_SLEEPTIME );
         if ( !IsTradeContextBusy() && IsTradeAllowed() ) break;
         else Print ( "====>       Trade context is busy... retry #" , i );
         if ( i == _BUSY_MAXTRY) {
            Print ( "====>       Trade context is busy... retry on next " + Symbol() + " tick");
            return ( false );
         }         
      }   

   for ( int n = 1; n <= _MODIFY_MAXTRY; n++ ) {         
      result = OrderModify( O_TICKET, NormalizeDigits( O_OPENPRICE ), NormalizeDigits( newSl ), NormalizeDigits( newTp ), CLR_NONE );
      // Select again the same order & refresh
      OrderSelect( O_TICKET, SELECT_BY_TICKET, MODE_TRADES );
      RefreshRates();
   
      if ( result ) {
         M_BID = MarketInfo( Symbol(), MODE_BID );
         M_ASK = MarketInfo( Symbol(), MODE_ASK );      
         O_STOPLOSS = OrderStopLoss();
         O_TAKEPROFIT = OrderTakeProfit();
         // =====> to Pipdefender.dll
         UpdateOrder( M_ASK, M_BID, O_STOPLOSS, O_TAKEPROFIT );      
         return ( true );
      }   
      else {      
         err = GetLastError();
         Print( "====>       ModifyOrder ERR=" + err + " | " , Symbol(), " #" + O_TICKET + " [" + OrderTypeToString() + "] | open=" + DoubleDigitsToString( O_OPENPRICE ) + " | Sl=", DoubleDigitsToString( OrderStopLoss() ) + " | newSl=" + DoubleDigitsToString(newSl) + " | " + OrderPriceName() + "=" + DoubleDigitsToString(OrderPrice()) );
         Print( ErrorDescription( err ) );
         switch( err ) {
            // Overcomable errors
            case 130 : // invalid stops            
               // this is a bad error... anyway it makes no sense to continue here... we must recalculate everything
               Sleep( _BUSY_SLEEPTIME );
               return ( false );
            case 135 : 
            case 136 : // no prices
               while( !RefreshRates() ) Sleep( 1 );                    
               continue;
            case 146 : // Trade context is busy
               Sleep( _BUSY_SLEEPTIME );
               continue;  
         } // end switch
         return ( false );
      } // end else
   } // end for
   return (false);
}

// Return TRUE only if the order was correctly closed, FALSE otherwise
bool CloseOrder() {
   bool result;
   int err;
     
   // Safety check for the the trade context...
   if( IsTradeContextBusy() || !IsTradeAllowed() ) {
      Print ( "====>       Trade context is busy... retry next tick..." );
      return ( false );
   }
   
   for ( int n = 1; n <= _CLOSE_MAXTRY; n++ ) {
      result = OrderClose( O_TICKET, OrderLots(), NormalizeDigits( OrderPrice() ), CLOSE_SLIPPAGE, CLR_NONE );      
      if ( !result ) {
         err = GetLastError();
         Print( "====>       ModifyOrder ERR=" + err + " | ", Symbol(), " #" + O_TICKET + " [" + OrderTypeToString() + "] | open=" + DoubleDigitsToString( O_OPENPRICE ) + " | " + OrderPriceName() + "=" + DoubleDigitsToString(OrderPrice()) );
         Print( ErrorDescription( err ) );
         switch( err ) {
            // Overcomable errors
            case 130 :
               M_BID = MarketInfo( Symbol(), MODE_BID );
               M_ASK = MarketInfo( Symbol(), MODE_ASK );
               continue;
            case 135 : 
            case 136 : // no prices
               while( !RefreshRates() ) Sleep( 1 );                    
               continue;
            case 146 : // Trade context is busy
               Sleep( _BUSY_SLEEPTIME ); 
               continue;  
         } // end switch
      } // end if     
      return ( result );           
   } // end for           
}


// -------------------------------------------------------------------------------------------------------------------------------
// LOGICAL FUNCTIONS
// -------------------------------------------------------------------------------------------------------------------------------

// Check the validity of the external parameters
void CheckParameters() {
   string errorString = ""; 
   
   // TRAILING_STOP
   if ( IsLesserEqual( TRAILING_STOP, X_STOPLEVEL ) )
      errorString = "The TRAILING_STOP must be greater than stop level ( > " + DoubleOneDigitToString( X_STOPLEVEL ) + " pip )";
   else
      // TRUE_SL
      if ( TRUE_SL < 2 * TRAILING_STOP )
         errorString = "The TRUE_SL must be be at least 2 * TRAILING_STOP";
      else
         // FAKE_TP
         if ( FAKE_TP < 2 * TRAILING_STOP )
            errorString = "The FAKE_TP must be at least 2 * TRAILING_STOP";   
         else
            // FAKE_SL
            if ( FAKE_SL < 2 * TRUE_SL ) 
               errorString = "The FAKE_SL must be at least 2 * TRUE_SL";
            else
               // MAX_HOURS_WARNING
               if ( ( MAX_HOURS_WARNING <=0 ) || ( MAX_HOURS_WARNING >= MAX_HOURS_ALARM ) )
                  errorString = "MAX_HOURS_WARNING must be greater than 0 and lesser than MAX_HOURS_ALARM";
               else
                  // MAX_HOURS_ALARM
                  if ( ( MAX_HOURS_ALARM <=0 ) || ( MAX_HOURS_ALARM <= MAX_HOURS_WARNING ) )
                     errorString = "MAX_HOURS_ALARM must be greater than 0 and greater than MAX_HOURS_WARNING";
                  else
                     // SAFE_LOSS
                     if ( SAFE_LOSS <= 0 )
                        errorString = "SAFE_LOSS must be greater than 0";
                     else
                        // GREEDY_LEVEL
                        if ( GREEDY_LEVEL <=0 )
                           errorString = "GREEDY_LEVEL must be greater than 0";
                  
   if ( errorString != "" ) {
      Alert( "====>       " + Symbol() + " parameter error! ", errorString );
      EXCEPTION = _ALERT;
   }
   else EXCEPTION = _ALL_OK;   
}

// -------------------------------------------------------------------------------------------------------------------------------
// UTILITY FUNCTIONS
// -------------------------------------------------------------------------------------------------------------------------------
void Cout(string text, bool flush=false) {
   static string textBuffer = "\n";
   
   textBuffer = textBuffer + _INDENT + text;
   Comment( textBuffer);
   if ( flush ) textBuffer = "\n";
}

void LineFeed() {
   Cout("");
}

// Cout & flush output
void CoutF( string text ) {
   Cout( text, true );
}

void ShowDisplay() {
   CoutF("");
   
   Cout( "Pipdefender           1.0beta ");
   Cout( "Beta expire date     31 April 2010" );
   Cout("------------------------------------");
   if ( FIBONACCI_MODE ) Cout("Fibonacci mode on");
   else Cout("Trailing Stop          " + DoubleOneDigitToString( TRAILING_STOP ) );
   
   Cout( "Stop Level             " + DoubleOneDigitToString( X_STOPLEVEL ) );
   Cout( "PointValue             " + DoubleDigitsToString( POINT_VALUE ));
   
   if ( !FIBONACCI_MODE ) 
      if ( GREEDY_MODE ) Cout( "Greedy mode         level #" + GREEDY_LEVEL + " on");
      else Cout( "Greedy mode            disabled");

   if ( STEALTH_MODE ) Cout( "Stealth mode on     " + DoubleOneDigitToString( TRUE_SL )  + " pip");
   else Cout( "Stealth mode disabled");

   if ( SAFE_MODE ) Cout( "Safe mode on        " + DoubleOneDigitToString( MAX_HOURS_WARNING )  + "/" + DoubleOneDigitToString( MAX_HOURS_ALARM )  + "h & " + DoubleOneDigitToString( SAFE_LOSS ));
   else Cout( "Safe mode off");

   Cout("------------------------------------");
}

double OrderPrice() {
   if ( O_TYPE == OP_BUY )  return ( M_BID );
   if ( O_TYPE == OP_SELL ) return ( M_ASK );
   Print("====>       OrderPrice() function: unknown order type.");
   EXCEPTION = _ABORT;
}

string OrderTypeToString(){
   if ( O_TYPE == OP_BUY )  return ( "buy" );
   if ( O_TYPE == OP_SELL ) return ( "sell" );
   Print("====>       OrderTypeToString() function: unknown order type.");
   EXCEPTION = _ABORT;
}

string OrderPriceName(){
   if ( O_TYPE == OP_BUY )  return ( "bid" );
   if ( O_TYPE == OP_SELL ) return ( "ask" );
   Print("====>       OrderPriceName() function: unknown order type.");
   EXCEPTION = _ABORT;
}

// The profit is a positive, negative or zero value
string ShowProfit() {
   if ( O_TYPE == OP_BUY )  return ( DoubleOneDigitToString( ( M_BID - O_OPENPRICE ) / POINT_VALUE ) + " pip" );
   if ( O_TYPE == OP_SELL ) return ( DoubleOneDigitToString( ( O_OPENPRICE - M_ASK ) / POINT_VALUE ) + " pip" );
   Print("====>       ShowProfit() function: unknown order type.");
   EXCEPTION = _ABORT;
}

